import { Add, Subtract } from "./arithmetic";
declare type _OmitByValue<O, V> = V extends infer Head | V ? never extends Head ? {
    [K in keyof O as V extends O[K] ? never : K]: O[K];
} : {
    [K in keyof O as Head extends O[K] ? never : K]: O[K];
} | _OmitByValue<O, Exclude<V, Head>> : never;
/**
 * Omit properties from an object whose value is in the given blacklist union
 */
declare type OmitByValue<O, Blacklist> = keyof _OmitByValue<O, Blacklist> extends never ? Record<string, never> : {
    [Key in keyof _OmitByValue<O, Blacklist>]: _OmitByValue<O, Blacklist>[Key];
};
/**
 * Removes any keys containing any of the types that have already been entered
 */
export declare type RemoveCirularReferences<T, Blacklist = T> = T extends [] ? T : T extends (infer M)[] ? (M extends object ? RemoveCirularReferences<Exclude<M, Blacklist>, Blacklist> : Exclude<M, Blacklist>)[] : T extends object ? {
    [K in keyof OmitByValue<T, T>]: RemoveCirularReferences<OmitByValue<T, T>[K], OmitByValue<T, T>[K] | Blacklist>;
} : T;
declare type DeepReqiuredWriteable<T> = {
    -readonly [P in keyof T]-?: DeepReqiuredWriteable<RemoveCirularReferences<NonNullable<T[P]>>>;
};
declare type KeyofArr<Arr extends [...unknown[]], I extends number = 1> = Arr extends [] ? never : Arr extends [unknown, ...infer Rest] ? 0 | Subtract<I, 1> | KeyofArr<Rest, Add<I, 1>> : number;
declare type KeyOf<T> = T extends object ? T extends Array<unknown> ? KeyofArr<T> : keyof T & string : never;
declare type MaybeSubtract<N extends number | undefined, Amount extends number> = N extends undefined ? undefined : Subtract<Exclude<N, undefined>, Amount>;
declare type _NestedKeyOf<T extends Record<string, unknown> | unknown[], MaxDepth extends number | undefined, Blacklist = never> = MaxDepth extends 0 ? KeyOf<T> : T extends Blacklist ? never : {
    [K in KeyOf<T>]: K | `${K}.${_NestedKeyOf<T[K], MaybeSubtract<MaxDepth, 1>>}`;
}[KeyOf<T>];
/**
 * Given an object with type {@link T}, returns a union of all paths to a value within {@link T}
 */
export declare type NestedKeyOf<T extends Record<string, unknown> | unknown[], MaxDepth extends number | undefined = undefined> = _NestedKeyOf<DeepReqiuredWriteable<T>, MaxDepth>;
declare type OptionalPropAccessor<T, K extends keyof Exclude<T, undefined>> = T extends undefined ? Exclude<T, undefined>[K] | undefined : Exclude<T, undefined>[K];
declare type _PathOf<Object, Path extends string | number> = Path extends `${infer Key}.${infer Rest}` ? Key extends keyof Exclude<Object, undefined> ? _PathOf<OptionalPropAccessor<Object, Key>, Rest> : never : Path extends keyof Exclude<Object, undefined> ? OptionalPropAccessor<Object, Path> : never;
/**
 * Given an object with type {@link T}, and a path with type {@link P}, returns the type of the value at {@link P} within {@link T}
 */
export declare type PathOf<T extends Record<string, unknown> | unknown[], P extends NestedKeyOf<T>> = _PathOf<T, P>;
export {};
